from gustaf.utils import arr, connec, log, tictoc
from gustaf.utils.tictoc import Tic

# Alias
connectivity = connec

__all__ = [
    "arr",
    "connec",
    "connectivity",
    "log",
    "tictoc",
    "Tic",
]
